HC = require 'HardonCollider'
require "TLfres"

local defines = require 'defines'
local classes = require "classes"
local Player = require "Player"
local StraightBullet = require "StraightBullet"
local Command = require "Command"
local Menu = require "Menu"

-- World
-- Todo... Get rid of these ugly globals.. Only need levelcommands, menu and state...
local g_topborder = nil
local g_leftborder = nil
local g_rightborder = nil
local g_bottomborder = nil

local g_playertopborder = nil
local g_playerleftborder = nil
local g_playerrightborder = nil
local g_playerbottomborder = nil

	  g_levelcommands = Command.new() --Trying to debug this, screw it I'm making it a global
local g_menu = Menu.new()

g_state = STATE_MENU

local g_width  = 1024
local g_height = 768-64

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-- Main Functions

function love.load()
	-- Setup collision
	Collider = HC(100, onCollide)

	-- Setup screen
	love.window.setMode(g_width, g_height)
	--TLfres.setScreen({g_width, g_height, false, false, 0}, g_width, true)
	love.mouse.setVisible(true)
	love.mouse.setGrabbed(true)

	-- Setup world
	g_world = love.physics.newWorld(0, 0, g_width, g_height)

	loadMenu()
end

function love.update(dt)
	Collider:update(dt)

	if (g_state == STATE_MENU) then
		g_menu:update(dt)
	elseif (g_state == STATE_GAME) then

		g_levelcommands:executeCommands(dt)

		g_levelcommands:update(dt)

		-- Check for clicking & create bullets
		if (love.mouse.isDown(1)) then
			if (g_levelcommands:checkFireTimer()) then
				g_levelcommands:firePlayerBullet()
			end
		end
	end

	Collider:update(dt)
end

function love.draw()
	--TLfres.transform()
	if (g_state == STATE_MENU) then
		g_menu:draw()
	elseif (g_state == STATE_GAME) then
		-- Debug info
		love.graphics.print("PlayerX: "..tostring(g_levelcommands.player:getX()).."\nPlayerY: "..tostring(g_levelcommands.player:getY()))
		--love.graphics.print()

		-- Update enemies / bullets / player
		g_levelcommands:draw()
	end
	--TLfres.letterbox(4,3)
end

function love.keypressed(key)
   --Debug
   if key == "f8" then
      debug.debug()
   end
	if key == "escape" then 
		love.event.quit()
	end
	if (g_state == STATE_MENU) then

	elseif (g_state == STATE_GAME) then
		if (key == "lshift") then
			g_levelcommands:setPlayerSpeed(SPEED_SLOW)
		end
	end
end

function love.keyreleased(key)
	if (g_state == STATE_MENU) then

	elseif (g_state == STATE_GAME) then
		if (key == "lshift") then
			g_levelcommands:setPlayerSpeed(SPEED_NORMAL)
		end
	end
end

function love.mousepressed(x, y, button)
	-- We actually check for a left click here (as opposed to in love.update) because we don't want it to repeatedly register when we're on the menu
	if (g_state == STATE_MENU) then
		if (button == 1) then
			g_menu:handleClick()
		end
	elseif (g_state == STATE_GAME) then
		if (button == 2) then
			g_levelcommands:setPlayerSpeed(SPEED_FAST)
		end
	end
end

function love.mousereleased(x, y, button)
	if (g_state == STATE_MENU) then

	elseif (g_state == STATE_GAME) then
		if (button == 2) then
			g_levelcommands:setPlayerSpeed(SPEED_NORMAL)
		end
	end
end

function beginGame()
	g_menu = nil
	Collider:clear()

	love.graphics.setColor(255, 255, 255)

	loadPlayer()

	g_levelcommands:updateMouse(love.mouse.getX(), love.mouse.getY())

	loadPower()

	createWalls()

	LoadLevel("levels/level2.txt")

	g_state = STATE_GAME
end

function createWalls() -- Note: get a better understanding of where these numbers come from
	-- Bullet borders
	g_topborder = Collider:addRectangle(-164, -164, g_width + 2 * 164, 100)
	g_topborder.type = WALL_TOP

	g_leftborder = Collider:addRectangle(-164, -63, 100, g_height + 228)
	g_leftborder.type = WALL_LEFT

	g_bottomborder = Collider:addRectangle(-63, g_height + 64, g_width + 228, 100)
	g_bottomborder.type = WALL_BOTTOM

	g_rightborder = Collider:addRectangle(g_width + 64, -63, 100, g_height + 125)
	g_rightborder.type = WALL_RIGHT

	-- Player border -- These are more important to get exact, player is directly affected by these
	g_playertopborder = Collider:addRectangle(-100, -100, g_width + 2 * 100, 100) -- last number was 50 for some reason
	g_playertopborder.type = WALL_PLAYERTOP

	g_playerleftborder = Collider:addRectangle(-100, 1, 100, g_width + 100)
	g_playerleftborder.type = WALL_PLAYERLEFT

	g_playerbottomborder = Collider:addRectangle(1, g_height, g_width + 100, 100)
	g_playerbottomborder.type = WALL_PLAYERBOTTOM

	g_playerrightborder = Collider:addRectangle(g_width, 1, 100, g_height - 2) -- Could have been a typo on their end, keeping the offset
	g_playerrightborder.type = WALL_PLAYERRIGHT

	--g_height = 768
	--g_width = 1024
	
	--Coordinates
	--Width Height

	-- top
	-- (-164, -164)
	-- 1352, 100

	--left
	--(-164, -63)
	--10, 996

	--bottom
	--(-63, 832)
	--1252, 100

	--right
	--(1088, 63)
	--100, 893
end

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-- Collision

function onCollide(dt, shape_a, shape_b)
	if (g_state == STATE_MENU) then
		local shape1 = collisionIsWall(shape_a)
		local shape2 = collisionIsWall(shape_b)

		if ((shape1 ~= WALL_NOT) or (shape2 ~= WALL_NOT)) then
			local wall = 0
			local other = 0

			if (shape1 ~= WALL_NOT) then
				wall = shape_a
				other = shape_b
			else
				wall = shape_b
				other = shape_a
			end

			if (other.type == TYPE_ENEMYBULLET) then
				g_menu:deleteBullet(other.index)
			end
		end
	elseif (g_state == STATE_GAME) then
		-- Check for wall collision first.. Will happen most frequently
		-- Wall = Outside the visible map, just used for collecting bullets

		local shape1 = collisionIsWall(shape_a)
		local shape2 = collisionIsWall(shape_b)

		if ((shape1 ~= WALL_NOT) or (shape2 ~= WALL_NOT)) then
			local wall = 0
			local other = 0

			if (shape1 ~= WALL_NOT) then
				wall = shape_a
				other = shape_b
			else
				wall = shape_b
				other = shape_a
			end

			-- React to the collision according to what type it is
			if (other.type == TYPE_PLAYERBULLET) then
				g_levelcommands:deletePlayerBullet(other.index)
			elseif (other.type == TYPE_ENEMYBULLET) then
				g_levelcommands:deleteBullet(other.index)
			end
			return
		end

		shape1 = collisionIsPlayerWall(shape_a)
		shape2 = collisionIsPlayerWall(shape_b)

		if ((shape1 ~= WALL_NOT) or (shape2 ~= WALL_NOT)) then
			local wall = 0
			local other = 0

			if (shape1 ~= WALL_NOT) then
				wall = shape_a
				other = shape_b
			else
				wall = shape_b
				other = shape_a
			end

			if (other.type == TYPE_PLAYER) then
				g_levelcommands:fixPositionWall(wall.type)
			end
			return
		end

		-- Collision with enemy
		shape1 = collisionIsEnemy(shape_a)
		shape2 = collisionIsEnemy(shape_b)

		if (shape1 or shape2) then
			local enemy
			local other

			if (shape1) then
				enemy = shape_a
				other = shape_b
			else
				enemy = shape_b
				other = shape_a
			end

			if (other.type == TYPE_PLAYER) then
				g_levelcommands:playerGetHit(dt)
			elseif (other.type == TYPE_PLAYERBULLET) then
				g_levelcommands:dealDamage(enemy.index)
				g_levelcommands:deletePlayerBullet(other.index)
			end
			return
		end

		-- Collision with Power
		shape1 = collisionIsPower(shape_a)
		shape2 = collisionIsPower(shape_b)

		if (shape1 or shape2) then
			local power
			local other

			if (shape1) then
				player = shape_a
				other = shape_b
			else
				player = shape_b
				other = shape_a
			end

			if (other.type == TYPE_PLAYER) then
				g_levelcommands:setPowerCollected(other.index)
			end
		end
	end
end

-- If shape is a wall, it will return the type of the wall
-- Otherwise, return NOTWALL
function collisionIsWall(shape)
	if ((shape == g_topborder) or (shape == g_rightborder) or (shape == g_leftborder) or (shape == g_bottomborder)) then
		return shape.type
	end
	return WALL_NOT
end

function collisionIsPlayerWall(shape)
	if ((shape == g_playertopborder) or (shape == g_playerrightborder) or (shape == g_playerleftborder) or (shape == g_playerbottomborder)) then
		return shape.type
	end
	return WALL_NOT
end

function collisionIsEnemy(shape)
	return ((shape.type == TYPE_ENEMY) or (shape.type == TYPE_ENEMYBULLET))
end

function collisionIsPower(shape)
	return (shape.type == TYPE_POWER)
end

function getPlayerID()
	return g_player
end

function removeEnemy(index)
	g_levelcommands:deleteEnemy(index)
end

function removePower(index)
	g_levelcommands:deletePower(index)
end

function createPower(x, y)
	g_levelcommands:createPower(x, y)
end

function increaseDamage(damage)
	g_levelcommands:increaseDamage(damage)
end

-------------------------------------------------------------------------------
-------------------------------------------------------------------------------
-- Loading

function loadMenu()
	g_menu:addCollision()

	local texture = "images/menu/quit.png"
	local texturehighlighted = "images/menu/quithighlighted.png"
	local x = 50
	local y = 620
	local size = 128
	local ysize = 64
	local buttontype = BUTTON_QUIT

	-- Load the defaults.. If it doesn't find the file then skip it
	exists = love.filesystem.exists("config/menu.cfg")
	if (exists) then
		for lines in love.filesystem.lines("config/menu.cfg") do
			if (lines == "") then
				if (buttontype == "QUIT") then
					g_menu:addButton(BUTTON_QUIT, texture, texturehighlighted, x, y, xsize, ysize)
				elseif (buttontype == "OPTIONS") then
					g_menu:addButton(BUTTON_OPTIONS, texture, texturehighlighted, x, y, xsize, ysize)
				elseif (buttontype == "PLAY") then
					g_menu:addButton(BUTTON_PLAY, texture, texturehighlighted, x, y, xsize, ysize)
				elseif (buttontype == "DECORATION") then
					g_menu:addButton(BUTTON_DECORATION, texture, texturehighlighted, x, y, xsize, ysize)
				end
			else
				local setting = split(lines, " ")

				if (setting[1] == "TEXTURE") then texture = setting[2]
				elseif (setting[1] == "TEXTUREHIGHLIGHTED") then texturehighlighted = setting[2]
				elseif (setting[1] == "X") then x = tonumber(setting[2])
				elseif (setting[1] == "Y") then y = tonumber(setting[2])
				elseif (setting[1] == "XSIZE") then xsize = tonumber(setting[2])
				elseif (setting[1] == "YSIZE") then ysize = tonumber(setting[2])
				elseif (setting[1] == "TYPE") then buttontype = setting[2]
				end
			end
		end
	end

	-- Load the animations for the menu
	exists = love.filesystem.exists("config/menuanimation.cfg")
	if (exists) then
		for lines in love.filesystem.lines("config/menuanimation.cfg") do
			g_menu:addCommand(lines)
		end
	end

	createWalls()

	g_menu:enableType(MENUSTATE_MAIN)
end

function loadPlayer()
	-- Incase the loading fails, we have defaults
	local startx = g_width / 2
	local starty = g_height - 20
	local xsize = 16
	local ysize = 16
	local slowspeed = 100
	local speed = 300
	local fastspeed = 500
	local texture = "images/regularboxes/playerbox.png"
	local bullettexture = "images/tinyboxes/bluebox.png"
	local damage = 1

	-- Load the defaults.. If it doesn't find the file then skip it
	exists = love.filesystem.exists("config/defaultplayer.cfg")
	if (exists) then
		for lines in love.filesystem.lines("config/defaultplayer.cfg") do
			local setting = split(lines, " ")

			if (setting[1] == "STARTX") then startx = tonumber(setting[2])
			elseif (setting[1] == "STARTY") then starty = tonumber(setting[2])
			elseif (setting[1] == "XSIZE") then xsize = tonumber(setting[2])
			elseif (setting[1] == "YSIZE") then ysize = tonumber(setting[2])
			elseif (setting[1] == "SLOWSPEED") then slowspeed = tonumber(setting[2])
			elseif (setting[1] == "SPEED") then speed = tonumber(setting[2])
			elseif (setting[1] == "FASTSPEED") then fastspeed = tonumber(setting[2])
			elseif (setting[1] == "TEXTURE") then texture = setting[2]
			elseif (setting[1] == "DAMAGE") then damage = tonumber(setting[2])
			elseif (setting[1] == "BULLETTEXTURE") then bullettexture = setting[2]
			end
		end
	end

	g_levelcommands:createPlayer(startx, starty, xsize, ysize, speed, texture, damage, bullettexture, slowspeed, fastspeed)
end

function loadPower()
	-- Incase the loading fails, we have defaults
	local stayuplength = 3
	local stayalivelength = 10
	local xsize = 16
	local ysize = 16
	local initialspeed = 0
	local dy = 5
	local texture = "images/regularboxes/powerbox.png"
	local texturecollected = "images/regularboxes/powerboxcollected.png"
	local poweramount = 0.5

	-- Load the defaults.. If it doesn't find the file then skip it
	exists = love.filesystem.exists("config/power.cfg")
	if (exists) then
		for lines in love.filesystem.lines("config/power.cfg") do
			local setting = split(lines, " ")

			if (setting[1] == "STAYUPLENGTH") then stayuplength = tonumber(setting[2])
			elseif (setting[1] == "STAYALIVELENGTH") then stayalivelength = tonumber(setting[2])
			elseif (setting[1] == "XSIZE") then xsize = tonumber(setting[2])
			elseif (setting[1] == "YSIZE") then ysize = tonumber(setting[2])
			elseif (setting[1] == "INITIALSPEED") then initialspeed = tonumber(setting[2])
			elseif (setting[1] == "DY") then dy = tonumber(setting[2])
			elseif (setting[1] == "TEXTURE") then texture = setting[2]
			elseif (setting[1] == "TEXTURECOLLECTED") then texturecollected = setting[2]
			elseif (setting[1] == "POWERAMOUNT") then poweramount = tonumber(setting[2])
			end
		end
	end

	g_levelcommands:setDefaultPower(stayuplength, stayalivelength, xsize, ysize, initialspeed, dy, texture, texturecollected, poweramount)
end

function LoadLevel(path)
	exists = love.filesystem.exists(path)
	if (exists) then
		--file = love.filesystem.newfile(path)
		--file:open('r')
		for lines in love.filesystem.lines(path) do
			g_levelcommands:addCommand(lines)
		end
	end
end